#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
shared utilities fo the my-snap tools
"""
# pylint: disable=broad-exception-caught,invalid-name

import re
import time
from datetime import datetime

##############################################################################
def timestamp_str():
    """ Get a data string for a time of so many days ago. """
    return datetime.now().strftime("%Y-%m-%d-%H%M%S")

##############################################################################
def human(number):
    """ Return a concise number description."""
    suffixes = ['K', 'M', 'G', 'T']
    number = float(number)
    while suffixes:
        suffix = suffixes.pop(0)
        number /= 1024
        if number < 99.95 or not suffixes:
            return f'{number:.1f}{suffix}'
    return None

##############################################################################
def ago_str(delta_secs, signed=False):
    """ Turn time differences in seconds to a compact representation;
        e.g., '18h·39m'
    """
    ago = int(max(0, round(delta_secs if delta_secs >= 0 else -delta_secs)))
    divs = (60, 60, 24, 7, 52, 9999999)
    units = ('s', 'm', 'h', 'd', 'w', 'y')
    vals = (ago%60, int(ago/60)) # seed with secs, mins (step til 2nd fits)
    uidx = 1 # best units
    for div in divs[1:]:
        # print('vals', vals, 'div', div)
        if vals[1] < div:
            break
        vals = (vals[1]%div, int(vals[1]/div))
        uidx += 1
    rv = '-' if signed and delta_secs < 0 else ''
    rv += f'{vals[1]}{units[uidx]}' if vals[1] else ''
    rv += f'{vals[0]:d}{units[uidx-1]}'
    return rv

##############################################################################
def ago_whence(filename):
    """ Find the standard time string in the file name and return
        the ago_str()
    """
    mat = re.search(r'\b(\d\d\d\d-\d\d-\d\d-\d\d\d\d\d\d)\b', filename)
    if mat:
        try:
            dt_object = datetime.strptime(mat.group(1), "%Y-%m-%d-%H%M%S")
            return ago_str(time.time() - dt_object.timestamp())
        except Exception:
            pass
    return ''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Custom Wrapper for python curses.
"""
# pylint: disable=too-many-instance-attributes,too-many-arguments
# pylint: disable=invalid-name,broad-except

import traceback
import atexit
import time
import curses
import textwrap
from types import SimpleNamespace
from curses.textpad import rectangle, Textbox
dump_str = None

class OptionSpinner:
    """Manage a bunch of options where the value is rotate thru
    a fixed set of values pressing a key."""
    def __init__(self):
        """Give the object with the attribute to change its
        value (e.g., options from argparse or "self" from
        the object managing the window).

        And array of specs like:
            ['a - allow auto suggestions', 'allow_auto', True, False],
            ['/ - filter pattern', 'filter_str', self.filter_str],
        A spec can have a trailing None + more comments shown after
        the value.
        """
        self.options, self.keys = [], []
        self.margin = 4 # + actual width (1st column right pos)
        self.align = self.margin # + actual width (1st column right pos)
        self.default_obj = SimpleNamespace() # if not given one
        self.attr_to_option = {} # given an attribute, find its option ns
        self.key_to_option = {} # given key, options namespace
        self.keys = set()

    @staticmethod
    def _make_option_ns():
        return SimpleNamespace(
            keys=[],
            descr='',
            obj=None,
            attr='',
            vals=None,
            prompt=None,
            comments=[],
        )

    def get_value(self, attr, coerce=False):
        """Get the value of the given attribute."""
        ns = self.attr_to_option.get(attr, None)
        obj = ns.obj if ns else None
        value = getattr(obj, attr, None) if obj else None
        if value is None and obj and coerce:
            if ns.vals:
                if value not in ns.vals:
                    value = ns.vals[0]
                    setattr(obj, attr, value)
            else:
                if value is None:
                    value = ''
                    setattr(ns.obj, ns.attr, '')
        return value

    def _register(self, ns):
        """ Create the mappings needed"""
        assert ns.attr not in self.attr_to_option
        self.attr_to_option[ns.attr] = ns
        for key in ns.keys:
            assert key not in self.key_to_option, f'key ({chr(key)}, {key}) already used'
            self.key_to_option[key] = ns
            self.keys.add(key)
        self.options.append(ns)
        self.align = max(self.align, self.margin+len(ns.descr))
        self.get_value(ns.attr, coerce=True)

    def add(self, obj, specs):
        """ Compatibility Method."""
        for spec in specs:
            ns = self._make_option_ns()
            ns.descr = spec[0]
            ns.obj = obj
            ns.attr = spec[1]
            ns.vals=spec[2:]
            if None in ns.vals:
                idx = ns.vals.index(None)
                ns.vals = ns.vals[:idx]
                ns.comments = ns.vals[idx+1:]
            ns.keys = [ord(ns.descr[0])]
            self._register(ns)

    def add_key(self, attr, descr, obj=None, vals=None, prompt=None, keys=None, comments=None):
        """ Standard method"""
        ns = self._make_option_ns()
        if keys:
            ns.keys = list(keys) if isinstance(keys, (list, tuple, set)) else [keys]
        else:
            ns.keys = [ord(descr[0])]
        if comments is None:
            ns.comments = []
        else:
            ns.comments = list(comments) if isinstance(keys, (list, tuple)) else [comments]
        ns.descr = descr
        ns.attr = attr
        ns.obj = obj if obj else self.default_obj
        ns.vals, ns.prompt = vals, prompt
        assert bool(ns.vals) ^ bool(ns.prompt)
        self._register(ns)

    @staticmethod
    def show_help_nav_keys(win):
        """For help screens, show the navigation keys. """
        for line in Window.get_nav_keys_blurb().splitlines():
            if line:
                win.add_header(line)

    def show_help_body(self, win):
        """ Write the help page section."""
        win.add_body('Type keys to alter choice:', curses.A_UNDERLINE)

        for ns in self.options:
            # get / coerce the current value
            value = self.get_value(ns.attr)
            assert value is not None, f'cannot get value of {repr(ns.attr)}'
            choices = ns.vals if ns.vals else [value]

            win.add_body(f'{ns.descr:>{self.align}}: ')

            for choice in choices:
                shown = f'{choice}'
                if isinstance(choice, bool):
                    shown = "ON" if choice else "off"
                win.add_body(' ', resume=True)
                win.add_body(shown, resume=True,
                    attr=curses.A_REVERSE if choice == value else None)

            for comment in ns.comments:
                win.add_body(f'{"":>{self.align}}:  {comment}')

    def do_key(self, key, win):
        """Do the automated processing of a key."""
        ns = self.key_to_option.get(key, None)
        if ns is None:
            return None
        value = self.get_value(ns.attr)
        if ns.vals:
            idx = ns.vals.index(value) if value in ns.vals else -1
            value = ns.vals[(idx+1) % len(ns.vals)] # choose next
        else:
            value = win.answer(prompt=ns.prompt, seed=value)
        setattr(ns.obj, ns.attr, value)
        return value

class Window:
    """ Layer above curses to encapsulate what we need """
    timeout_ms = 200
    static_scr = None
    nav_keys = """
        Navigation:    H/M/L:   top/middle/end-of-page
            k, UP:  up one row               0, HOME:  first row
          j, DOWN:  down one row              $, END:  last row
           Ctrl-u:  half-page up       Ctrl-b, PPAGE:  page up
           Ctrl-d:  half-page down     Ctrl-f, NPAGE:  page down
    """
    def __init__(self, head_line=True, head_rows=50, body_rows=200,
                 body_cols=200, keys=None, pick_mode=False, pick_size=1):
        self.scr = self._start_curses()

        self.head = SimpleNamespace(
            pad=curses.newpad(head_rows, body_cols),
            rows=head_rows,
            cols=body_cols,
            row_cnt=0,  # no. head rows added
            texts = [],
            view_cnt=0,  # no. head rows viewable (NOT in body)
        )
        self.body = SimpleNamespace(
            pad = curses.newpad(body_rows, body_cols),
            rows= body_rows,
            cols=body_cols,
            row_cnt = 0,
            texts = []
        )
        self.hor_line_cnt = 1 if head_line else 0 # no. h-lines in header
        self.scroll_pos = 0  # how far down into body are we?
        self.max_scroll_pos = 0
        self.pick_pos = 0 # in highlight mode, where are we?
        self.last_pick_pos = -1 # last highlighted position
        self.pick_mode = pick_mode # whether in highlight mode
        self.pick_size = pick_size # whether in highlight mode
        self.rows, self.cols = 0, 0
        self.scroll_view_size = 0  # no. viewable lines of the body
        self.handled_keys = set(keys) if isinstance(keys, (set, list)) else []
        self._set_screen_dims()
        self.calc()

    @staticmethod
    def get_nav_keys_blurb():
        """For a help screen, describe the nav keys"""
        return textwrap.dedent(Window.nav_keys)

    def _set_screen_dims(self):
        """Recalculate dimensions ... return True if geometry changed."""
        rows, cols = self.scr.getmaxyx()
        same = bool(rows == self.rows and cols == self.cols)
        self.rows, self.cols = rows, cols
        return same

    @staticmethod
    def _start_curses():
        """ Curses initial setup.  Note: not using curses.wrapper because we
        don't wish to change the colors. """
        atexit.register(Window.stop_curses)
        Window.static_scr = scr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        curses.curs_set(0)
        scr.keypad(1)
        scr.timeout(Window.timeout_ms)
        scr.clear()
        return scr

    def set_pick_mode(self, on=True, pick_size=1):
        """Set whether in highlight mode."""
        was_on, was_size = self.pick_mode, self.pick_size
        self.pick_mode = bool(on)
        self.pick_size = max(pick_size, 1)
        if self.pick_mode and (not was_on or was_size != self.pick_size):
            self.last_pick_pos = -2 # indicates need to clear them all

    @staticmethod
    def stop_curses():
        """ Curses shutdown (registered to be called on exit). """
        if Window.static_scr:
            curses.nocbreak()
            curses.echo()
            Window.static_scr.keypad(0)
            curses.endwin()
            Window.static_scr = None

    def calc(self):
        """Recalculate dimensions ... return True if geometry changed."""
        same = self._set_screen_dims()
        self.head.view_cnt = min(self.rows - self.hor_line_cnt, self.head.row_cnt)
        self.scroll_view_size = self.rows - self.head.view_cnt - self.hor_line_cnt
        self.max_scroll_pos = max(self.body.row_cnt - self.scroll_view_size, 0)
        self.body_base = self.head.view_cnt + self.hor_line_cnt
        return not same

    def _put(self, ns, *args):
        """ Add text to head/body pad using its namespace. args:
            - bytes (converted to str/unicode)
            - str (unicode)
            - None (same as curses.A_NORMAL)
            - int (curses attribute)
        """
        def flush(attr=None):
            nonlocal self, is_body, row, text, seg, first
            if (is_body and self.pick_mode) or attr is None:
                attr = curses.A_NORMAL
            if seg and first:
                ns.pad.addstr(row, 0, seg, attr)
            elif seg:
                ns.pad.addstr(seg, attr)
            text += seg
            seg, first, attr = '', False, None

        is_body = bool(id(ns) == id(self.body))
        if ns.row_cnt < ns.rows:
            row = max(ns.row_cnt, 0)
            text, seg, first = '', '', True
            for arg in args:
                if isinstance(arg, bytes):
                    arg = arg.decode('utf-8')
                if isinstance(arg, str):
                    seg += arg  # note: add w/o spacing
                elif arg is None or isinstance(arg, (int)):
                    # assume arg is attribute ... flushes text
                    flush(attr=arg)
            flush()
            ns.texts.append(text)  # text only history
            ns.row_cnt += 1

    def put_head(self, *args):
        """ Put a line above the line."""
        self._put(self.head, *args)

    def put_body(self, *args):
        """ Put a line below the line."""
        self._put(self.body, *args)

    def _add(self, ns, text, attr=None, resume=False):
        """ Add text to head/body pad using its namespace"""
        is_body = bool(id(ns) == id(self.body))
        if ns.row_cnt < ns.rows:
            row = max(ns.row_cnt - (1 if resume else 0), 0)
            if (is_body and self.pick_mode) or attr is None:
                attr = curses.A_NORMAL
            if resume:
                ns.pad.addstr(text, attr)
                ns.texts[row] += text
            else:
                ns.pad.addstr(row, 0, text, attr)
                ns.texts.append(text)  # text only history
                ns.row_cnt += 1

    def add_header(self, text, attr=None, resume=False):
        """Add text to header"""
        self._add(self.head, text, attr, resume)

    def add_body(self, text, attr=None, resume=False):
        """ Add text to body (below header and header line)"""
        self._add(self.body, text, attr, resume)

    def draw(self, y, x, text, text_attr=None, width=None, leftpad=False, header=False):
        """Draws the given text (as utf-8 or unicode) at position (row=y,col=x)
        with optional text attributes and width.
        This is more compatible with my older, simpler Window class.
        """
        ns = self.head if header else self.body
        text_attr = text_attr if text_attr else curses.A_NORMAL
        if y < 0 or y >= ns.rows or x < 0 or x >= ns.cols:
            return # nada if out of bounds
        if y+1 >= ns.row_cnt:
            ns.row_cnt = y+1

        uni = text if isinstance(text, str) else text.decode('utf-8')

        if width is not None:
            width = min(width, self.cols - x)
            if width <= 0:
                return
            padlen = width - len(uni)
            if padlen > 0:
                if leftpad:
                    uni = padlen * ' ' + uni
                else:  # rightpad
                    uni += padlen * ' '
            text = uni[:width].encode('utf-8')
        else:
            text = uni.encode('utf-8')

        try:
            while y >= len(ns.texts):
                ns.texts.append('')
            ns.texts[y] = ns.texts[y][:x].ljust(x) + uni + ns.texts[y][x+len(uni):]
            ns.pad.addstr(y, x, text, text_attr)
        except curses.error:
            # this sucks, but curses returns an error if drawing the last character
            # on the screen always.  this can happen if resizing screen even if
            # special care is taken.  So, we just ignore errors.  Anyhow, you cannot
            # get decent error handling.
            pass


    def highlight_picked(self):
        """Highlight the current pick and un-highlight the previous pick."""
        def get_text(pos):
            nonlocal self
            return self.body.texts[pos] if pos < len(self.body.texts) else ''

        if not self.pick_mode:
            return
        pos0, pos1 = self.last_pick_pos, self.pick_pos
        if pos0 == -2: # special flag to clear all formatting
            for row in range(self.body.row_cnt):
                self.body.pad.addstr(row, 0, get_text(row), curses.A_NORMAL)
        if pos0 != pos1:
            if 0 <= pos0 < self.body.row_cnt:
                for i in range(self.pick_size):
                    self.body.pad.addstr(pos0+i, 0, get_text(pos0+i), curses.A_NORMAL)
            if 0 <= pos1 < self.body.row_cnt:
                for i in range(self.pick_size):
                    string = get_text(pos1+i)
                    self.body.pad.addstr(pos1+i, 0, string, curses.A_REVERSE)
                self.last_pick_pos = pos1

    def _scroll_indicator_row(self):
        """ Compute the absolute scroll indicator row:
        - We want the top to be only when scroll_pos==0
        - We want the bottom to be only when scroll_pos=max_scroll_pos-1
        """
        if self.max_scroll_pos <= 1:
            return self.body_base
        y2, y1 = self.scroll_view_size-1, 1
        x2, x1 = self.max_scroll_pos, 1
        x = self.scroll_pos
        pos = y1 + (y2-y1)*(x-x1)/(x2-x1)
        return min(self.body_base + int(max(pos, 0)), self.rows-1)

    def _scroll_indicator_col(self):
        """ Compute the absolute scroll indicator col:
        - We want the left to be only when scroll_pos==0
        - We want the right to be only when scroll_pos=max_scroll_pos-1
        """
        if self.pick_mode:
            return self._calc_indicator(
                self.pick_pos, 0, self.body.row_cnt-1, 0, self.cols-1)
        return self._calc_indicator(
            self.scroll_pos, 0, self.max_scroll_pos, 0, self.cols-1)

    def _calc_indicator(self, pos, pos0, pos9, ind0, ind9):
        if self.max_scroll_pos <= 0:
            return -1 # not scrollable
        if pos9 - pos0 <= 0:
            return -1 # not scrollable
        if pos <= pos0:
            return ind0
        if pos >= pos9:
            return ind9
        ind = int(round(ind0 + (ind9-ind0+1)*(pos-pos0)/(pos9-pos0+1)))
        return min(max(ind, ind0+1), ind9-1)

    def render(self):
        """Draw everything added. In a loop cuz curses is a
        piece of shit."""
        for _ in range(128):
            try:
                self.render_once()
                return
            except curses.error:
                time.sleep(0.16)
                self._set_screen_dims()
                continue
        try:
            self.render_once()
        except curses.error:
            Window.stop_curses()
            print(f"""curses err:
    head.row_cnt={self.head.row_cnt}
    head.view_cnt={self.head.view_cnt}
    hor_line_cnt={self.hor_line_cnt}
    body.row_cnt={self.body.row_cnt}
    scroll_pos={self.scroll_pos}
    max_scroll_pos={self.max_scroll_pos}
    pick_pos={self.pick_pos}
    last_pick_pos={self.last_pick_pos}
    pick_mode={self.pick_mode}
    pick_size={self.pick_size}
    rows={self.rows}
    cols={self.cols}
""")
            raise


    def render_once(self):
        """Draw everything added."""
        self.calc()
        # if self.scroll_view_size <= 0:
            # self.scr.refresh()
        indent = 0
        if self.body_base < self.rows:
            ind_pos = 0 if self.pick_mode else self._scroll_indicator_row()
            if self.pick_mode:
                self.pick_pos = max(self.pick_pos, 0)
                self.pick_pos = min(self.pick_pos, self.body.row_cnt-1)
                if self.pick_pos >= 0:
                    self.pick_pos -= (self.pick_pos % self.pick_size)
                if self.pick_pos < 0:
                    self.scroll_pos = 0
                elif self.scroll_pos > self.pick_pos:
                    # light position is below body bottom
                    self.scroll_pos = self.pick_pos
                elif self.scroll_pos < self.pick_pos - (self.scroll_view_size - self.pick_size):
                    # light position is above body top
                    self.scroll_pos = self.pick_pos - (self.scroll_view_size - self.pick_size)
                self.scroll_pos = max(self.scroll_pos, 0)
                self.scroll_pos = min(self.scroll_pos, self.max_scroll_pos)
                indent = 1
            else:
                self.scroll_pos = max(self.scroll_pos, 0)
                self.scroll_pos = min(self.scroll_pos, self.max_scroll_pos)
                self.pick_pos = self.scroll_pos + ind_pos - self.body_base
                # indent = 1 if self.body.row_cnt > self.scroll_view_size else 0

        if indent > 0 and self.pick_mode:
            self.scr.vline(self.body_base, 0, ' ', self.scroll_view_size)
            if self.pick_pos >= 0:
                pos = self.pick_pos - self.scroll_pos + self.body_base
                self.scr.addstr(pos, 0, '>', curses.A_REVERSE)

        if self.head.view_cnt < self.rows:
            self.scr.hline(self.head.view_cnt, 0, curses.ACS_HLINE, self.cols)
            ind_pos = self._scroll_indicator_col()
            if ind_pos >= 0:
                bot, cnt = ind_pos, 1
                if 0 < ind_pos < self.cols-1:
                    width = (self.scroll_view_size/self.body.row_cnt*self.cols)
                    bot = max(int(round(ind_pos-width/2)), 1)
                    top = min(int(round(ind_pos+width/2)), self.cols-1)
                    cnt = top - bot
                # self.scr.addstr(self.head.view_cnt, bot, '-'*cnt, curses.A_REVERSE)
                # self.scr.hline(self.head.view_cnt, bot, curses.ACS_HLINE, curses.A_REVERSE, cnt)
                for idx in range(bot, bot+cnt):
                    self.scr.addch(self.head.view_cnt, idx, curses.ACS_HLINE, curses.A_REVERSE)

        self.scr.refresh()

        if self.rows > 0:
            last_row = min(self.head.view_cnt, self.rows)-1
            if last_row >= 0:
                self.head.pad.refresh(0, 0, 0, indent, last_row, self.cols-1)

        if self.body_base < self.rows:
            if self.pick_mode:
                self.highlight_picked()
            self.body.pad.refresh(self.scroll_pos, 0,
                  self.body_base, indent, self.rows-1, self.cols-1)


    def answer(self, prompt='Type string [then Enter]', seed='', width=80):
        """Popup"""
        def mod_key(key):
            return  7 if key == 10 else key

        # need 3 extra cols for rectangle (so we don't draw in southeast corner)
        # and 3 rows (top/prompt/bottom)
        #      +Prompt---    -----------------+
        #      | Seed-for-answer              |
        #      +---------Press ENTER to submit+
        if self.rows < 3 or self.cols < 30:
            return seed
        width = min(width, self.cols-3) # max text width
        row0, row9 = self.rows//2 - 1, self.rows//2 + 1
        col0 = (self.cols - (width+2)) // 2
        col9 = col0 + width + 2 - 1

        self.scr.clear()
        win = curses.newwin(1, width, row0+1, col0+1) # input window
        rectangle(self.scr, row0, col0, row9, col9)
        self.scr.addstr(row0, col0+1, prompt[0:width])
        win.addstr(seed[0:width-1])
        ending = 'Press ENTER to submit'[:width]
        self.scr.addstr(row9, col0+1+width-len(ending), ending)
        self.scr.refresh()
        curses.curs_set(2)
        answer = Textbox(win).edit(mod_key).strip()
        curses.curs_set(0)
        return answer

    def alert(self, title='ALERT', message='', height=1, width=80):
        """Alert box"""
        def mod_key(key):
            return  7 if key == 10 else key

        # need 3 extra cols for rectangle (so we don't draw in southeast corner)
        # and 3 rows (top/prompt/bottom)
        #      +Prompt---    -----------------+
        #      | First line for message...    |
        #      | Last line for message.       |
        #      +-----------Press ENTER  to ack+
        if self.rows < 2+height or self.cols < 30:
            return
        width = min(width, self.cols-3) # max text width
        row0 = (self.rows+height-1)//2 - 1
        row9 = row0 + height + 1
        col0 = (self.cols - (width+2)) // 2
        col9 = col0 + width + 2 - 1

        self.scr.clear()
        for row in range(self.rows):
            self.scr.insstr(row, 0, ' '*self.cols, curses.A_REVERSE)
        pad = curses.newpad(20, 200)
        win = curses.newwin(1, 1, row9-1, col9-2) # input window
        rectangle(self.scr, row0, col0, row9, col9)
        self.scr.addstr(row0, col0+1, title[0:width], curses.A_REVERSE)
        pad.addstr(message)
        ending = 'Press ENTER to ack'[:width]
        self.scr.addstr(row9, col0+1+width-len(ending), ending)
        self.scr.refresh()
        pad.refresh(0, 0, row0+1, col0+1, row9-1, col9-1)
        Textbox(win).edit(mod_key).strip()
        return

    def clear(self):
        """Clear in prep for new screen"""
        self.scr.clear()
        self.head.pad.clear()
        self.body.pad.clear()
        self.head.texts, self.body.texts, self.last_pick_pos = [], [], -1
        self.head.row_cnt = self.body.row_cnt = 0

    def prompt(self, seconds=1.0):
        """Here is where we sleep waiting for commands or timeout"""
        ctl_b, ctl_d, ctl_f, ctl_u = 2, 4, 6, 21
        elapsed = 0.0
        while elapsed < seconds:
            key = self.scr.getch()
            if key == curses.ERR:
                elapsed += self.timeout_ms / 1000
                continue
            if key in (curses.KEY_RESIZE, ) or curses.is_term_resized(self.rows, self.cols):
                # self.scr.erase()
                self._set_screen_dims()
                # self.render()
                break

            # App keys...
            if key in self.handled_keys:
                return key # return for handling

            # Navigation Keys...
            pos = self.pick_pos if self.pick_mode else self.scroll_pos
            delta = self.pick_size if self.pick_mode else 1
            was_pos = pos
            if key in (ord('k'), curses.KEY_UP):
                pos -= delta
            elif key in (ord('j'), curses.KEY_DOWN):
                pos += delta
            elif key in (ctl_b, curses.KEY_PPAGE):
                pos -= self.scroll_view_size
            elif key in (ctl_u, ):
                pos -= self.scroll_view_size//2
            elif key in (ctl_f, curses.KEY_NPAGE):
                pos += self.scroll_view_size
            elif key in (ctl_d, ):
                pos += self.scroll_view_size//2
            elif key in (ord('0'), curses.KEY_HOME):
                pos = 0
            elif key in (ord('$'), curses.KEY_END):
                pos = self.body.row_cnt - 1
            elif key in (ord('H'), ):
                pos = self.scroll_pos
            elif key in (ord('M'), ):
                pos = self.scroll_pos + self.scroll_view_size//2
            elif key in (ord('L'), ):
                pos = self.scroll_pos + self.scroll_view_size-1

            if self.pick_mode:
                self.pick_pos = pos
            else:
                self.scroll_pos = pos
                self.pick_pos = pos

            if pos != was_pos:
                self.render()
            # ignore unhandled keys
        return None

def no_runner():
    """Appease sbrun"""

if __name__ == '__main__':
    def main():
        """Test program"""
        def do_key(key):
            nonlocal spin, win, opts
            value = spin.do_key(key, win)
            if key in (ord('p'), ord('s')):
                win.set_pick_mode(on=opts.pick_mode, pick_size=opts.pick_size)
            elif key == ord('n'):
                win.alert(title='Info', message=f'got: {value}')
            return value

        spin = OptionSpinner()
        spin.add_key('help_mode', '? - toggle help screen', vals=[False, True])
        spin.add_key('pick_mode', 'p - toggle pick mode', vals=[False, True])
        spin.add_key('pick_size', 's - #rows in pick', vals=[1, 2, 3])
        spin.add_key('name', 'n - select name', prompt='Provide Your Name:')
        spin.add_key('mult', 'm - row multiplier', vals=[0.5, 0.9, 1.0, 1.1, 2, 4, 16])
        opts = spin.default_obj

        win = Window(head_line=True, keys=spin.keys)
        opts.name = "[hit 'n' to enter name]"
        for loop in range(100000000000):
            body_size = int(round(win.scroll_view_size*opts.mult))
            if opts.help_mode:
                win.set_pick_mode(False)
                spin.show_help_nav_keys(win)
                spin.show_help_body(win)
            else:
                win.set_pick_mode(opts.pick_mode, opts.pick_size)
                win.add_header(f'Header: {loop} "{opts.name}"')
                for idx, line in enumerate(range(body_size//opts.pick_size)):
                    win.add_body(f'Main pick: {loop}.{line}')
                    for num in range(1, opts.pick_size):
                        win.draw(num+idx*opts.pick_size, 0, f'  addon: {loop}.{line}')
            win.render()
            _ = do_key(win.prompt(seconds=5))
            win.clear()

    try:
        main()
    except KeyboardInterrupt:
        Window.stop_curses()
        print(" === Quitting due to KeyboardInterrupt")
    except Exception as exce:
        Window.stop_curses()
        print("exception:", str(exce))
        print(traceback.format_exc())
        if dump_str:
            print(dump_str)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
BTRFS snapshot tool implementing the simplest snapshot strategy; i.e.,
  * replace eldest snapshots with new ones before updates, and
  * generally keep those snapshots until the next update.
  
NOTE: for basic debugging, pass --DB and it will dump major objects w/o
starting the window.
"""
# pylint: disable=invalid-name,consider-using-with,too-few-public-methods
# pylint: disable=redefined-outer-name,consider-using-generator
# pylint: disable=too-many-locals,broad-exception-caught,multiple-statements
# pylint: disable=too-many-instance-attributes,import-outside-toplevel
# pylint: disable=global-statement,broad-exception-raised,too-many-branches
# pylint: disable=too-many-statements,too-many-arguments

import re
import os
import sys
import atexit
import traceback
import subprocess
import curses as cs
from types import SimpleNamespace
try:
    from PowerWindow import Window, OptionSpinner
    from MyUtils import human, ago_whence, timestamp_str
except:
    from my_snaps.PowerWindow import Window, OptionSpinner
    from my_snaps.MyUtils import human, ago_whence, timestamp_str

##############################################################################

class BTRFS:
    """ TBD """
    def __init__(self, opts):
        self.tmp_dir = '/tmp/.btrfs/'
        self.temps = {} # keyed by device (e.g., /dev/nvme0n1p2)
        self.devs = {}
        self.snap_subvol = None
        self.DB = opts.DB
        self.add_limit = opts.add_snap_max
        self.label = None
        self.help_mode = False
        self.win = None
        self.rows = []
        self.dirty = True # TBD: need to refresh knowledge
        self.show_size = False # until we calc disk usage

        self.blkid_lines = [] # to avoid rerunning "blkid" on refresh
        self.mounts_lines = [] # to avoid rereading "/proc/mounts" on refresh

        atexit.register(self.umount_tmps)

        if opts.cron:
            self._install_cron_job(opts)
            sys.exit(0)

        self._refresh_if_dirty()

        if self.DB:
            self._get_disk_usage()
            sys.exit(0)


        if self.add_limit > 0:
            if opts.label:
                self.label = '=' + opts.label.replace('=', '')
            success = self._replace_eldest_snaps()
            print("OK" if success else "FAIL", f'add_snap_limit={self.add_limit}')
            if success and opts.print:
                self._refresh_if_dirty()
                self._print()
            sys.exit(0 if success else 1)

        if opts.print:
            self._print()
            sys.exit(0)

        self._start_window()

    def _print(self):
        for dev in self.devs.values():
            print(f'df: {dev.diskfree}')
        path_width = self.calc_path_width()
        mounts_width = self.calc_mounts_width()
        devs_width = self.calc_devs_width()
        for row in self.rows:
            wds = row.path.split('/.snapshots/', maxsplit=1)
            shown_path = f'  | {wds[1]}' if len(wds) > 1 else row.path
            mount_str = row.mount if row.mount else '' if row.subvol_ns.snap_of else '~'
            print(
                  f'{mount_str:>{mounts_width}}'
                  f' {row.dev:>{devs_width}}'
                  f' {shown_path:<{path_width}}')

    def _install_cron_job(self, opts):
        """ Add/replace an anacron job for scheduled snapshots """
        dirname = f'/etc/cron.{opts.cron}'
        if not os.path.isdir(dirname):
            print('ERROR: {dirname!r} does not exist')
            sys.exit(-1)
        filename = os.path.join(dirname, f'{opts.cron}-snaps')
        text = '#!/bin/sh\n'
        text += f'{sys.executable} {os.path.abspath(__file__)} -p -s{opts.add_snap_max}'
        text += f' -L{opts.label} >/tmp/.my-snaps-{opts.cron}.txt 2>&1\n'
        with open(filename, mode='w', encoding='utf-8') as f:
            f.write(text)
        os.chmod(filename, 0o755)
        print(f'OK: to {filename!r}, wrote:\n{text}')

    def _refresh_if_dirty(self):
        if not self.dirty:
            return
        # save sizes for xfer to refresh (cuz expensive)
        sizes = {}
        for ns in self.subvol_iter():
            sizes[f'{ns.dev}{ns.path}'] = ns.size
        self._load_devs()  # learn all the BTRFS partitions
        self._mount_tmps()  # mount BTRFS partitions in /tmp/.btrfs
        self._determine_mount_points() # associated current mounts with subvols
        self.gather_snapshots()  # determine subvols that are snapshots
        # restore sizes after refresh
        for ns in self.subvol_iter():
            key = f'{ns.dev}{ns.path}'
            if key in sizes:
                ns.size = sizes[key]
        self.make_rows()
        self.dirty = False

    def subvol_iter(self, subvol_ns=None, top_down=True):
        """  subvolume iterator recursively top-down (or bottom-up)
         - by default, all the subvols, not just top-level ones
         - optionally, just one tree given a subvol
        """
        def children_iter(subvol_ns, top_down):
            if top_down:
                yield subvol_ns
            for ns in subvol_ns.children:
                yield from children_iter(ns, top_down)
            if not top_down:
                yield subvol_ns

        if subvol_ns:
            yield from children_iter(subvol_ns, top_down=top_down)
            return

        for dev_ns in self.devs.values():
            # pylint: disable=redefined-argument-from-local
            for subvol_ns in dev_ns.subvols:
                if subvol_ns.depth == 0:
                    yield from children_iter(subvol_ns, top_down=top_down)

    def _cur_snap_suffix(self):
        return '.' + timestamp_str() + ('' if self.label is None else self.label)

    def stop_curses(self):
        """Terminate curses if running."""
        if self.win:
            self.win.stop_curses()

    def _start_window(self):
        def do_key(key):
            nonlocal spin, win, self
            value = spin.do_key(key, win)
            if key in (ord('u'), ) and not self.help_mode:
                win.clear()
                win.set_pick_mode(False)
                self.refresh_info(body=' ... BE REALLY PATIENT (running "du") ...')
                win.render()
                self._get_disk_usage()
                win.set_pick_mode(True)
                self.refresh_info()

            elif key in (ord('r'), ) and not self.help_mode:
                self._replace_eldest_snaps()

            elif key in (ord('a'), ) and not self.help_mode:
                self._replace_eldest_snaps(just_add=True)

            elif key in (ord('s'), ) and not self.help_mode:
                self._create_snap()

            elif key in (ord('d'), ) and not self.help_mode:
                self._del_subvolume()

            elif key in (ord('x'), ) and not self.help_mode:
                self.stop_curses()
                print('\n   OK, QUITTING NOW\n')
                sys.exit(0)

            elif key in (cs.KEY_ENTER, 10) and self.help_mode:
                self.help_mode = False
                win.set_pick_mode(False)

            elif key == ord('n'):
                win.alert(title='Info', message=f'got: {value}')

            return value

        spin = OptionSpinner()
        spin.add_key('help_mode', '? - toggle help screen', vals=[False, True], obj=self)

        base_keys_we_handle=[cs.KEY_ENTER,
                10, ord('s'), ord('d'), ord('u'), ord('r'), ord('a'), ord('x')]

        win = self.win = Window(keys=set(list(spin.keys) + list(base_keys_we_handle)))

        for _ in range(100000000000):
            if self.help_mode:
                win.set_pick_mode(False)
                spin.show_help_nav_keys(win)
                spin.show_help_body(win)
                self.win.add_body('Action keys:', attr=cs.A_UNDERLINE)
                self.win.add_body(' d - delete highlighted item')
                self.win.add_body(' s - create snapshot for highlighted item')
                self.win.add_body(' u - compute "du" for all subvols (very slow)')
                self.win.add_body(' r - replace eldest snapshot of each subvol')
                self.win.add_body(' a - add snapshot o each subvol with snapshots')
                self.win.add_body(' x - exit')
            else:
                win.set_pick_mode(True)
                self.refresh_info()
            win.render()
            try:
                _ = do_key(win.prompt(seconds=300))
            except Exception as exce:
                win.stop_curses()
                print("exception:", str(exce))
                print(traceback.format_exc())
                sys.exit(15)
            win.clear()

    def _replace_eldest_snaps_of_subvol(self, subvol_ns, like_snaps,
                                        ans=None, discard=1):
        """ For ONE subvolume having snapshots, remove those and create a new snapshot."""

        if not subvol_ns.snaps:
            return True

        while discard > 0 and like_snaps:
            if not self._del_subvolume(like_snaps[0], ans="y"):
                return False  # all must succeed
            subvol_ns.snaps.remove(like_snaps[0])
            like_snaps.pop(0)
            discard -= 1
        return self._create_snap(subvol_ns, ans=ans)


    def _replace_eldest_snaps(self, discard=1, just_add=False):
        """ TBD """
        ans, success = None, None

        if self.add_limit > 0:
            ans = self._cur_snap_suffix()

        for _ in (0, 1):
            labels = set()
            for dev_ns in self.devs.values():
                # pylint: disable=redefined-argument-from-local
                for subvol_ns in dev_ns.subvols:
                    if subvol_ns == self.snap_subvol:
                        continue
                    like_snaps =  []
                    if subvol_ns.depth != 0 or not subvol_ns.mount:
                        continue
                    for snap in subvol_ns.snaps:
                        if self.label is None or snap.snap_label == self.label:
                            like_snaps.append(snap)
                            labels.add(snap.snap_label if snap.snap_label else "''")
                    if ans or self.add_limit > 0:
                        this_cnt = len(like_snaps)
                        if self.add_limit > 0:
                            discard = max(0, this_cnt-self.add_limit+1)
                        elif just_add:
                            discard = max(0, this_cnt-8+1)

                        rv = self._replace_eldest_snaps_of_subvol(subvol_ns,
                                         like_snaps, ans=ans, discard=discard)
                        success = rv if success is None else rv if success else False
                        self.label = None
            if ans:
                break
            while True:
                ans = self.win.answer(
                        f'Set snap suffix OR clear; labels: {",".join(list(labels))}',
                            seed=ans if ans else self._cur_snap_suffix())
                if ans and (len(ans) < 5 or ans[:1] != '.'
                            or any(x in ans[1:] for x in './')):
                    self.win.alert(
                        'answer must be "." PLUS 4 or more characters w/o any "." or "/"')
                else:
                    self.label = self._get_label(ans)
                    break
            if not ans:
                break
        return success


    def _create_snap(self, subvol_ns=None, ans=None):
        if not subvol_ns:
            subvol_ns = self.rows[self.win.pick_pos].subvol_ns
        dev_ns = self.devs[subvol_ns.dev]

        if subvol_ns.snap_of:
            self.win.alert('Sorry, cannot create snapshot of snapshot')
            return False

        if not subvol_ns.mount:
            self.win.alert('Sorry, cannot create snapshot of unmounted subvolume')
            return False

        if subvol_ns.mount == '/.snapshots':
            self.win.alert('Sorry, cannot create snapshot of snapshot subvolume')
            return False

        if not ans:
            seed=self._cur_snap_suffix()
            ans = self.win.answer(
                f'Set suffix for snap "{subvol_ns.path}" OR clear', seed=seed)
        if not ans:
            return False

        snap_dir = f'{dev_ns.tmp_path}{self.snap_subvol.path}'
        snap_path = f'{snap_dir}{subvol_ns.path}{ans}'

        cmd = f'btrfs sub snap -r {subvol_ns.mount} {snap_path}'
        out, err, code = self._slurp_command(cmd)
        if code:
            self.win.alert(f'FAILED({code}): {cmd}', message='\n'.join(out + err),
                           height=len(out)+len(err))
            return False
        self.dirty = True
        return True

    def _del_subvolume(self, subvol_ns=None, ans=""):
        if not subvol_ns:
            subvol_ns = self.rows[self.win.pick_pos].subvol_ns

        for ns in self.subvol_iter(subvol_ns, top_down=False):
            if ns.mount:
                self.win.alert(f'Sorry, cannot delete mounted subvol {ns.mount}')
                return False

        if not ans:
            ans = self.win.answer(f'Type "y" to Delete "{subvol_ns.path}"')
        if not ans.strip().lower().startswith('y'):
            return False

        dev_ns = self.devs[subvol_ns.dev]
        for ns in self.subvol_iter(subvol_ns, top_down=False):
            snap_path = f'{dev_ns.tmp_path}{ns.path}'
            cmd = f'btrfs sub del {snap_path}'
            out, err, code = self._slurp_command(cmd)
            if code:
                self.win.alert('FAILED: {cmd}', message='\n'.join(out + err),
                               height=len(out)+len(err))
                return False
            self.dirty = True
        return True

    def _slurp_command(self, command):
        if self.DB: print('DB: +', command)
        process = subprocess.Popen(command, stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE, shell=True)
        status = process.wait()
        output, err = process.communicate()
        output, err = output.decode('utf-8'), err.decode('utf-8')
        output, err = output.splitlines(keepends=False), err.splitlines(keepends=False)
        return (output, err, os.WEXITSTATUS(status))

    @staticmethod
    def _slurp_file(pathname):
        with open(pathname, "r", encoding='utf-8') as fh:
            return [line.strip() for line in fh]

    @staticmethod
    def dev_path(dev):
        """ Return the full path to a dev (aka device basename) """
        return f'/dev/{dev}'

    def _load_devs(self):
        """ Discover all the BTRS Devices """

        if not self.blkid_lines:
            self.blkid_lines, _, _ = self._slurp_command('blkid')
        self.devs = {}
        for line in self.blkid_lines:
            # /dev/nvme0n1p2: LABEL="btrfs-common"
            #   UUID="8f60fc2f-872d-4327-aff9-34c4c4cefde7"
            #   UUID_SUB="d7b0987a-1133-4844-a19b-c6c22350379a"
            #   BLOCK_SIZE="4096" TYPE="btrfs"
            #   PARTUUID="02b5122d-5229-c347-a351-142008b89149"

            matches = re.findall(r'(\w+)="([^"]+)"', line)
            ns = SimpleNamespace() # Create a dictionary to store the fields and values
            for match in matches:
                field, value = match[0].lower(), match[1]
                if field in ('type',):
                    setattr(ns, field, value)
            if not hasattr(ns, 'type'):
                continue
            ns.dev = os.path.basename(line.split(': ', maxsplit=1)[0])
            if ns.type in ('btrfs', ) and ns.dev:
                delattr(ns, 'type')
                rows, _, _ = self._slurp_command(f'df -h {self.dev_path(ns.dev)}')
                ns.diskfree = rows[1] if len(rows) >= 2 else ''
                self.devs[ns.dev] = ns
        if self.DB:
            print('DB: --->>> after load_devs()')
            for dev, ns in self.devs.items():
                print(f'DB: {dev}: {vars(ns)}')

    @staticmethod
    def init_subvol_ns(dev='', path='', ident=None, parent=None):
        """ Create a subvolume namespace"""
        return SimpleNamespace(dev=dev, path=path,
               size=None, depth=0, mount='', snaps=[],
                children=[], snap_label='', snap_of=None,
                ident=ident, parent=parent, ago_str='')

    def _mount_tmps(self):
        """ mount each btrfs as needed """
        os.makedirs(self.tmp_dir, exist_ok=True)
        for dev, dev_ns in self.devs.items():
            tmp_mount_dir = os.path.join(self.tmp_dir, dev)
            if not os.path.ismount(tmp_mount_dir):
                os.makedirs(tmp_mount_dir, exist_ok=True)
                code = os.WEXITSTATUS(os.system(
                        f'set -x; mount {self.dev_path(dev_ns.dev)} {tmp_mount_dir}'))
                if code:
                    raise Exception(f'cannot mount {dev_ns.dev} {code=}')
            dev_ns.tmp_path = tmp_mount_dir
            lines, _, _ = self._slurp_command(f'btrfs sub list {tmp_mount_dir}')

            dev_ns.paths = {} # subvols keyed by relative path
            dev_ns.idents = {} # subvols keyed by ident
            dev_ns.subvols = []
            for line in lines:
                wds = line.split(maxsplit=8)
                if len(wds) < 9:
                    continue
                ident, parent, path = wds[1], wds[6], f'/{wds[8]}'
                child = BTRFS.init_subvol_ns(dev=dev,
                             path=path, ident=ident, parent=parent)
                if parent in dev_ns.idents:
                    parent_ns = dev_ns.idents[parent]
                    child.depth = parent_ns.depth+1
                    parent_ns.children.append(child)
                else: # top-levels are in the dev
                    dev_ns.subvols.append(child)
                dev_ns.idents[ident] = child
                dev_ns.paths[path] = child
                child.ago_str = ago_whence(child.path)

        if self.DB:
            print('DB: --->>> after mount_tmps()')
            for dev, dev_ns in self.devs.items():
                print(f'DB: {dev=}: keys={vars(dev_ns).keys()}')
            for subvol in self.subvol_iter():
                print(f'DB:   {"  "*subvol.depth}subvol: {vars(subvol)}')


    def umount_tmps(self):
        """ unmount each btrfs as needed """
        for ns in self.devs.values():
            os.system(f'set -x; umount {ns.tmp_path}')

    def _determine_mount_points(self):
        if not self.mounts_lines:
            self.mounts_lines = self._slurp_file('/proc/mounts')
        for line in self.mounts_lines:
            wds = re.split(r'\s+', line)
            if len(wds) < 4:
                continue

            mount, fstype, opts = wds[1], wds[2], wds[3]
            if mount == '/' and fstype != 'btrfs':
                assert False, f'root is not BTRFS ({mount}, {fstype})'
            if fstype != 'btrfs':
                continue

            match = re.search(r'subvolid=(\d+)', opts)
            if not match:
                continue
            ident = match.group(1)
            for subvol in self.subvol_iter():
                if subvol.ident == ident:
                    subvol.mount = mount
                    if mount == '/.snapshots':
                        self.snap_subvol = subvol
                    break
        for dev_ns in self.devs.values():
            dev_ns.subvols = sorted(dev_ns.subvols,
                    key=lambda x: (x.mount if x.mount else '/~~~~~~~~~~~~', x.path))

        if self.DB:
            print('DB: --->>> after determine_mount_points()')
            for subvol in self.subvol_iter():
                if subvol.mount:
                    print(f'DB:   {subvol.dev=} {subvol.mount=} {subvol.path=}')
        assert self.snap_subvol, "cannot find mounted /.snapshots"


    @staticmethod
    def _get_label(string):
        wds = string.rsplit('=', maxsplit=1)
        if len(wds) > 1:
            return f'={wds[1]}'
        return ''

    def gather_snapshots(self):
        """ TBD """
        def link(of_subvol, subvol):
            if of_subvol and subvol:
                of_subvol.snaps.append(subvol)
                of_subvol.snaps = sorted(of_subvol.snaps, key=lambda x: x.path)
                subvol.snap_of = of_subvol
                subvol.snap_label = self._get_label(remainder)

        for subvol in self.subvol_iter():
            wds = subvol.path.split('@snapshots/', maxsplit=1)
            if len(wds) < 2:
                continue
            mat = re.match(r'^(.*)\.[\-\d\:]+', wds[1])
            if not mat:
                link(self.snap_subvol, subvol)
                continue
            of_path = f'/{mat.group(1)}'
            remainder = wds[1][len(of_path):] # everything after the .
            dev_ns = self.devs[subvol.dev]
            of_subvol = dev_ns.paths.get(of_path, None)
            if of_subvol:
                link(of_subvol, subvol)
            else:
                link(self.snap_subvol, subvol)

        if self.DB:
            print('DB: --->>> after gather_snapshots()')
            for subvol in self.subvol_iter():
                if subvol.snaps:
                    print(f'DB:   {subvol.dev=} {subvol.mount=} {subvol.path=}')
                for snap in subvol.snaps:
                    print(f'DB:   snap {snap.path=}')


    def make_rows(self):
        """ Create the set of rows for display with only the subset of info
        needed for display"""
        def init_row(size, mount, dev, path, subvol_ns):
            return SimpleNamespace(size=size, mount=mount,
                               dev=dev, path=path, subvol_ns=subvol_ns)

        self.rows = []

        for dev, dev_ns in self.devs.items():
            for ns in dev_ns.subvols:
                row = init_row(ns.size, ns.mount, dev, ns.path, ns)
                self.rows.append(row)
                for snap in ns.snaps:
                    row = init_row(snap.size,
                           snap.mount, snap.dev, snap.path, snap)
                    self.rows.append(row)
        if self.DB:
            for row in self.rows:
                print(f'DB: row: {row.size=} {row.mount=}'
                      f' {row.dev=} {row.path=!r}')

    def _get_disk_usage(self):
        """This actually only works for the snaps."""
        def convert_human(val):
            if val.endswith('GiB'):
                val = float(val[:-3]) * (2**30)
            elif val.endswith('MiB'):
                val = float(val[:-3]) * (2**20)
            elif val.endswith('KiB'):
                val = float(val[:-3]) * (2**10)
            elif re.search(r'\dB$', val):
                val = float(val[:-1]) * 1
            else:
                print(f'ERR: cannot parse({val}')
                val = 0
            return int(round(val))

        if not self.snap_subvol:
            return
        dev_ns = self.devs.get(self.snap_subvol.dev, None)
        if not dev_ns:
            return

        cmd = f'cd "{dev_ns.tmp_path}" && btrfs fi du -s {self.snap_subvol.path[1:]}/*'
        lines, _, _ = self._slurp_command(cmd)
        for line in lines[1:]:
            total, exclusive, _, pathname = re.split(r'\s+', line.strip(), maxsplit=3)
            pathname = os.path.join('/', pathname)
            if not pathname in dev_ns.paths:
                continue
            snap_ns = dev_ns.paths[pathname]
            snap_ns.size = convert_human(exclusive)
            if snap_ns.snap_of:
                if not snap_ns.snap_of.size:
                    snap_ns.snap_of.size = 0
                snap_ns.snap_of.size = max(convert_human(total),
                    snap_ns.snap_of.size)

        for row in self.rows:
            row.size = row.subvol_ns.size
        self.show_size = True

        if self.DB:
            print('DB: --->>> after _get_disk_usage()')
            for ns in self.subvol_iter():
                print(f'DB: {ns.size=} {ns.path=!r}')

    def calc_path_width(self):
        """ TBD """
        rv = max([len(x.path) for x in self.rows])
        return max(len('Subvolume'), rv)

    def calc_mounts_width(self):
        """ TBD """
        rv = max([len(x.mount) for x in self.rows])
        return max(len('Mount'), rv)

    def calc_devs_width(self):
        """ TBD """
        rv = max([len(x.dev) for x in self.rows])
        return max(len('Device'), rv)

    def refresh_info(self, body=None):
        """ TBD """

        win = self.win
        self._refresh_if_dirty()
        path_width = self.calc_path_width()
        mounts_width = self.calc_mounts_width()
        devs_width = self.calc_devs_width()

        win.add_header('MY-SNAPS', attr=cs.A_REVERSE)
        win.add_header(
            '  s:+snap d:-subvol u:disk-usage r:replace-all a:add-all x:exit ?:help',
            resume=True)
        for dev_ns in self.devs.values():
            win.add_header(f'df: {dev_ns.diskfree}')
        size_hdr = f' {"~Size":>7}'
        win.add_header(
              f'{"Mount":>{mounts_width}}'
              f'{size_hdr if self.show_size else ""}'
              f' {"Device":>{devs_width}}'
              f' {"Subvolume":<{path_width}}',
              attr=cs.A_BOLD,
              )

        if body:
            body = body if isinstance(body, (list, tuple)) else [body]
            for line in body:
                win.add_body(line, attr=cs.A_REVERSE)
            return

        for row in self.rows:
            wds = row.path.split('@snapshots/', maxsplit=1)
            if len(wds) < 2:
                shown_path = row.path
            else:
                shown_path = f'--> {wds[1]}'
                if row.subvol_ns.ago_str:
                    shown_path += f' {row.subvol_ns.ago_str}'
            if row.subvol_ns.snap_of == self.snap_subvol:
                shown_path = '!!!' + shown_path[3:]
            mount_str = row.mount if row.mount else '' if row.subvol_ns.snap_of else '~'
            size_str = f' {"-" if row.size is None else human(row.size):>7}'
            win.add_body(
                  f'{mount_str:>{mounts_width}}'
                  f'{size_str if self.show_size else ""}'
                  f' {row.dev:>{devs_width}}'
                  f' {shown_path:<{path_width}}')

btrfs = None
def main():
    """ TBD """
    global btrfs
    import argparse
    if os.geteuid() != 0: # Re-run the script with sudo
        os.execvp('sudo', ['sudo', sys.executable] + sys.argv)

    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--add-snap-max', type=int, default=0,
            help='add snapshots limited to value per subvol [1<=val<=8]')
    parser.add_argument('-L', '--label', type=str,
            help='add given label to -s snapshots')
    parser.add_argument('-p', '--print', action="store_true",
            help='print the subvolumes/snaps and exit')
    parser.add_argument('--cron', type=str,
            choices=('hourly', 'daily', 'weekly', 'monthly'),
            help='install a periodic snapshot anacron job')
    parser.add_argument('--DB', action="store_true",
            help='add some debugging output')
    opts = parser.parse_args()
    if opts.add_snap_max > 0:
        opts.add_snap_max = min(opts.add_snap_max, 8)
    if opts.cron:
        if not opts.label:
            opts.label = '=' + opts.cron.capitalize()
        if not opts.add_snap_max:
            defaults={'hourly': 2, 'daily': 4, 'weekly': 2, 'monthly':1}
            opts.add_snap_max = defaults[opts.cron]
        opts.print = False

    btrfs = BTRFS(opts)
    btrfs.umount_tmps()

def run():
    try:
        main()
    except KeyboardInterrupt:
        if btrfs and btrfs.win:
            btrfs.stop_curses()
        print('\n   OK, QUITTING NOW\n')
        sys.exit(0)
    except Exception as exce:
        if btrfs and btrfs.win:
            btrfs.stop_curses()
        print("exception:", str(exce))
        print(traceback.format_exc())
        sys.exit(15)

if __name__ == "__main__":
    run()


#!/usr/bin/env python3
""" Program to run """
# pylint: disable=invalid-name,broad-exception-caught
# pylint: disable=too-many-locals,too-many-arguments

import sys
import os
import subprocess
import re
import traceback
from types import SimpleNamespace
try:
    from InlineMenu import Menu
    from MyUtils import timestamp_str, ago_whence
except:
    from my_snaps.InlineMenu import Menu
    from my_snaps.MyUtils import timestamp_str, ago_whence

class BtrfsRestore:
    """ TBD """
    def __init__(self):
        if os.geteuid() != 0: # Re-run the script with sudo
            print('NOTE: restarting with "sudo"')
            os.execvp('sudo', ['sudo', sys.executable] + sys.argv)
        self.dry_run = False
        self.filesystems = []
        self.slash_mnt = self.get_slash_mnt()

    def do_command(self, prompts, todo=None, precmd='', once=False, force=False):
        """ TBD"""
        prompts['x'] = 'EXIT'
        while True:
            prompts['y'] = ('toggle dry-run='
                            + ('ON' if self.dry_run else 'OFF'))
            menu = Menu(prompts, str(todo) if todo else None,
                        title='my-restore menu'
                          + (' DRY-RUN' if self.dry_run else ''))
            choice = menu.get_prompt_obj()
            if choice.key == 'x':
                sys.exit(0)
            if choice.key == 'y':
                self.dry_run = not self.dry_run
                continue
            cmd = menu.get_command(choice)
            if not cmd.startswith('#'):
                if not force and self.dry_run:
                    os.system(f'echo WOULD + {precmd!r} {cmd!r}')
                else:
                    os.system(f'clear;set -x; {precmd} {cmd}')
            todo = choice.next
            if once:
                return todo

    def get_slash_mnt(self):
        """ TBD """
        def slurp_file(pathname):
            with open(pathname, "r", encoding='utf-8') as fh:
                return [line.strip() for line in fh]

        rv = SimpleNamespace(device='', fstype='')
        lines = slurp_file('/proc/mounts')
        for line in lines:
            wds = re.split(r'\s+', line)
            if len(wds) >= 4:
                device, mount, fstype, = wds[0], wds[1], wds[2]
                if mount == '/mnt':
                    rv = SimpleNamespace(device=device, fstype=fstype)
        return rv

    def select_mount(self):
        """ TBD """
        command_output = subprocess.check_output(['btrfs', 'filesystem', 'show']).decode('utf-8')
        lines = command_output.split('\n')
        self.filesystems = []
        for line in lines:
            mat = re.match(r"^Label:\s+('[^']*')", line)
            if mat:
                label = mat.group(1)
            mat = re.search(r"\bpath\s+(/dev/\S+)", line)
            if mat:
                path = mat.group(1)
                self.filesystems.append(SimpleNamespace(label=label, path=path))
        cmds, todo, mnt = {}, '-', self.slash_mnt
        for idx, fs in enumerate(self.filesystems):
            if fs.path == mnt.device:
                cmd = f'# KEEP {fs.path!r} mounted on /mnt'
            else:
                cmd = 'umount /dev && ' if mnt.device else ''
                cmd += f'mount {fs.path} /mnt # {fs.label}'
            cmds[str(idx)] = cmd
            todo = '0'
        todo = self.do_command(cmds, todo, once=True, force=True)

    def get_state(self):
        """ Create a dict of subvolumes that have a snapshots and/or a reverted tip """
        command_output = subprocess.check_output([
            'btrfs', 'sub', 'list', '.']).decode('utf-8')
        lines = command_output.split('\n')
        # ID 667 gen 216849 top level 5 path eos@my-opt
        # ID 699 gen 217076 top level 5 path eos@snapshots
        # ID 782 ... eos@snapshots/eos@root.2024-01-10-174732=Update
        # ID 791 ... eos@snapshots/eos@root.2024-01-13-084102=Daily
        subnames = set()
        subs = {}
        reverts = {}
        for line in lines:
            mat = re.search(r"\bpath\s+(\S+)", line)
            if not mat:
                continue
            subpath = mat.group(1)
            basename = os.path.basename(subpath)
            parent = os.path.dirname(subpath)
            if not parent and not basename.endswith('@snapshot'):
                if '.' in basename and basename.endswith('=Reverted'):
                    reverts[basename.split('.', 1)[0]] = subpath
                elif not '.' in basename:
                    subnames.add(basename)
                continue
            if parent.endswith('@snapshots'):
                wds = basename.split('.', 1)
                if len(wds) > 1:
                    subvol = wds[0]
                    sub_snaps = subs.get(subvol, [])
                    sub_snaps.append(subpath)
                    subs[subvol] = sub_snaps
        rv = {}
        for name in sorted(subnames):
            sub_snaps = subs.get(name, [])
            revert = reverts.get(name, None)
            if sub_snaps or revert:
                rv[name] = SimpleNamespace(
                    revert=revert, snaps=sorted(sub_snaps))

        # for name, ns in rv.items(): print(f'{name}: {ns}')
        return rv

    def select_restores(self, todo='a'):
        """ TBD """
        def advance(key, lead):
            return Menu.get_next_key(key), ' '*len(lead)

        # pylint: disable=too-many-branches
        os.chdir('/mnt')
        subs = self.get_state()
        key, cmds = 'a', {}

        for subvol, ns in subs.items():
            lead = f'{subvol}'
            if ns.revert:
                cmds[key] = [
                    f'{lead}: revert {ns.revert}',
                    f'btrfs sub del "{subvol}" && mv "{ns.revert}" "{subvol}"',
                ]
                key, lead = advance(key, lead)
                cmds[key] = [
                    f'{lead}: del {ns.revert}',
                    f'btrfs sub del "{ns.revert}"',
                ]
                key, lead = advance(key, lead)
                prep = f'btrfs sub del "{subvol}" &&'
            else:
                prep = f'mv "{subvol}" "{subvol}.{timestamp_str()}=Reverted" &&'

            for snap in ns.snaps:
                snap_base = os.path.basename(snap)
                cmds[key] = [
                    f'{lead}: restore {snap_base} {ago_whence(snap_base)}',
                    f'{prep} btrfs sub snap "{snap}" "{subvol}"',
                ]
                key, lead = advance(key, lead)
        cmds[key] = 'reboot now'
        return self.do_command(cmds, once=True, todo=todo)

    def main(self):
        """ The top-level function. """
        # pylint: disable=import-outside-toplevel
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument('-n', '--dry-run', action="store_true",
                help='do NOT do anything')
        opts = parser.parse_args()
        self.dry_run = opts.dry_run
        self.select_mount()
        os.system('clear')
        todo = 'a'
        while True:
            todo = self.select_restores(todo=todo)


def run():
    """Wrap main in try/except."""
    try:
        BtrfsRestore().main()
    except KeyboardInterrupt:
        pass
    except Exception as exce:
        print("exception:", str(exce))
        print(traceback.format_exc())


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
""" A simple menu tool w/o the full curses ado.
Create the menu by passing in:
- an optional 'default' key where the menu starts
    - if default is an object with attribute 'next', that will be use
- an optional 'title'
- 'prompts' which is a dict of the form {key: prompt, ...} where:
    - 'key' is one character (typically abc...012..ABC...)
        - often 'x' is reserved for exiting the menu if looped
    - each 'prompt' may be:
        - a single line string
        - a multiple line string
            - the first line is the actual prompt
            - the remainder is called extra
        - list or tuple of values:
            - the first element is a single or multiple line string (as above)
            - if no more elements, 'payload' is None
            - if one more element, 'payload' is prompt[1]
            - if more than one element, 'payload' is prompt[1:]

- Menu.prompt() just returns the key
- Menu.get_prompt_obj() return a SimpleNamespace with members:
    - key: the key selected
    - prompt: the prompt shown
    - extra: described above (None of only single line given)
    - payload: described above when prompt is a list or tuple
    - next: the next key
"""
# pylint: disable=invalid-name,broad-except,too-many-instance-attributes
# pylint: disable=too-few-public-methods

import sys
import termios
import tty
import shutil
import signal
from types import SimpleNamespace

####################################################################################

class Term:
    """ Escape sequences; e.g., see:
     - https://en.wikipedia.org/wiki/ANSI_escape_code
     - https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797#file-ansi-md
    """
    esc = '\x1B'
    # pylint: disable=missing-function-docstring,multiple-statements
    @staticmethod
    def erase_line(): return f'{Term.esc}[2K'
    @staticmethod
    def bold(): return f'{Term.esc}[1m'
    @staticmethod
    def reverse_video(): return f'{Term.esc}[7m'
    @staticmethod
    def normal_video(): return f'{Term.esc}[m'
    @staticmethod
    def pos_up(cnt): return f'{Term.esc}[{cnt}F' if cnt > 0 else ''
    @staticmethod
    def pos_down(cnt): return f'{Term.esc}[{cnt}E' if cnt > 0 else ''
    @staticmethod
    def col(pos): return f'{Term.esc}[{pos}G'
    @staticmethod
    def clear_screen(): return f'{Term.esc}[H{Term.esc}[2J{Term.esc}[3J'

####################################################################################

class Menu:
    """ A simple menu system """
    fd = sys.stdin.fileno()
    save_attrs = None

    class InputTimeout(Exception):
        """ Raised for obvious reasons. """
        
    @staticmethod
    def _prompt_mode(enable=False):
        """ Used to start / complete prompting the user so that we can read
            one character at a time, etc.
        """
        if enable:
            if not Menu.save_attrs:
                Menu.save_attrs = termios.tcgetattr(Menu.fd)
                tty.setcbreak(Menu.fd)
        else:
            if Menu.save_attrs:
                termios.tcsetattr(Menu.fd, termios.TCSADRAIN, Menu.save_attrs)
                Menu.save_attrs = None

    def __init__(self, prompts, default=None, title=''):
        self.prompts = prompts
        self.lines, self.keys = [], []
        self.objects = {}
        self.selected = len(prompts)
        self.title = title
        self.cols, self.rows = shutil.get_terminal_size()
        self.max_line = 0
        # time.sleep(10)
        self.pos = 0
        self._parse_prompts()
        self.input = ''
        self.default_bottom = 'Use Up/Down/key to highlight and Enter to select'
        self.prev_bottom = self.default_bottom
        self._refresh(clear=False)
        self.pos = len(self.lines)
        if hasattr(default, 'next'):
            default = default.next
        default = default if default else self.keys[0]
        if default in self.prompts:
            idx = list(self.keys).index(default)
            self._move(idx - self.selected)

    def _parse_prompts(self):
        for key, prompt in self.prompts.items():
            payload, extra = None, None
            if key in self.objects:
                print(f'WARNING: skip dup key ({key})')
                continue
            if isinstance(prompt, (tuple, list)):
                prompt, payload = prompt[0], prompt[1:]
                payload = (None if not payload else
                           payload[0] if len(payload) == 1 else payload)
            lines = prompt.splitlines()
            if len(lines) > 1:
                prompt, extra = lines[0], '\n'.join(lines[1:])
            prompt = prompt.rstrip()
            self.lines.append(f'{key}: {prompt}')
            self.keys.append(key)
            self.objects[key] = SimpleNamespace(key=key, prompt=prompt,
                                   extra=extra, payload=payload, next=None)
        for idx, ns in enumerate(self.objects.values()):
            ns.next = self.keys[ min(idx+1, len(self.keys)-1) ]

    def _refresh(self, clear=True):
        """ Draw/redraw the menu optionally clearing the screen.
            Do this a first and after screen resizes.
        """
        self.pos = len(self.prompts)
        if clear:
            print(Term.clear_screen(), end='', flush=True)
        print(f'\n\n     {Term.bold()}<<<< {self.title} >>>>'
              + Term.normal_video() + '\n'*len(self.prompts))
        for idx in range(0, len(self.lines)+1):
            print(self._get_line_str(idx), end='', flush=True)
        self._bottom_line()
        print(self._set_pos_str(self.selected), end='', flush=True)

    def _set_pos_str(self, idx):
        """ Goto the given line from where we are at."""
        idx = 0 if idx < 0 else len(self.lines) if idx > len(self.lines) else idx
        if idx == self.pos:
            return ''
        if idx < self.pos:
            self.pos, cnt = idx, self.pos - idx
            return Term.pos_up(cnt)
        self.pos, cnt = idx, idx - self.pos
        return Term.pos_down(cnt)

    def _get_line_str(self, idx):
        """ Returns the string to write for the given line of the menu.
            'where' overrides the line to put it on.
        """
        idx = 0 if idx < 0 else len(self.lines) if idx > len(self.lines) else idx
        action, pre, on, off = '', ' ', '', ''
        if idx == self.selected and idx < len(self.lines):
            pre, on, off = '>',Term.reverse_video(), Term.normal_video()
        if idx < len(self.lines):
            self.max_line = max(2+len(self.lines[idx]), self.max_line)
            action += self._set_pos_str(idx)
            action += Term.erase_line()
            line = self.lines[idx][:self.cols-2]
            action += f'{pre} {on}{line}{off}'
            action += Term.col(0)
        return action

    def _move(self, cnt):
        """ Move the cursor by the cnt lines up (negative) or down (positive).
            If actually moving, then the old cursor will be un-highlighted
            and the new cursor is highlighted.
        """
        action = ''
        idx = self.selected + cnt
        idx = 0 if idx < 0 else len(self.lines)-1 if idx >= len(self.lines) else idx
        if idx != self.selected:
            was_selected, self.selected = self.selected, idx
            action += self._get_line_str(was_selected)
            action += self._get_line_str(self.selected)
            action += self._set_pos_str(self.selected)
            print(action, end='', flush=True)

    def _get_char(self):
        """ Get one char from stdin """
        def init_timeout_handling():
            # `SIGWINCH` is send on terminal resizes
            def handle_timeout(*ignore):
                raise Menu.InputTimeout
            signal.signal(signal.SIGALRM, handle_timeout)
            signal.setitimer(signal.ITIMER_REAL, 0.25, 0.25)

        def reset_timeout_handling() -> None:
            signal.setitimer(signal.ITIMER_REAL, 0, 0)
            signal.signal(signal.SIGALRM, signal.SIG_IGN)

        rv = None
        init_timeout_handling()
        while True:
            try:
                init_timeout_handling()
                rv = sys.stdin.read(1)
                reset_timeout_handling()
            except Menu.InputTimeout:
                reset_timeout_handling()
            except Exception as caught:
                reset_timeout_handling()
                raise caught

            reset_timeout_handling()
            cols, rows = shutil.get_terminal_size()
            if self.cols != cols or self.rows != rows:
                self.cols, self.rows = cols, rows
                # self._bottom_line(f'resize: rows={self.rows}, cols={self.cols}')
                self._bottom_line()
                self._refresh()
            if rv is not None:
                return rv

    def _bottom_line(self, string=None):
        """ Draw/redraw the bottom line possibly with an error message """
        self.prev_bottom = string if string else self.prev_bottom
        action = self._set_pos_str(len(self.lines))
        action += Term.erase_line()
        bottom = f':::: {self.prev_bottom} ::::'
        action += bottom[:self.cols-1]
        action += self._set_pos_str(self.selected)
        action += Term.col(0)
        print(action, end='', flush=True)

    def _finish(self, string=''):
        """ Complete the menu selection by moving the cursor after the
            menu and printing any given message.
        """
#       cnt = 1 + len(self.lines) - self.selected
#       action = Term.pos_down(cnt) + string
#       print(action, flush=True)
#       for _ in range(cnt-1, -1):
#           action += Term.pos_up(1) + Term.erase_line()
        # action += '    ' + Term.reverse_video()
        # self.max_line = max(len(self.lines[self.selected]), self.max_line)
        # action += self.lines[self.selected][:self.cols]
#       print(action, 'RUN\n', flush=True)
        for idx in range(len(self.lines)):
            print(self._set_pos_str(idx) + Term.erase_line(), end='')
        print(Term.pos_down(1) + Term.erase_line(), end='', flush=True)
        action = Term.pos_up(1+len(self.lines)) + Term.erase_line()
        action += Term.reverse_video() + 'PICK' + Term.normal_video() + ': '
        action += self.lines[self.selected][:self.cols-6]
        print(action + Term.pos_down(3), end='', flush=True)

    def _restore_default_bottom(self):
        """ After the user does something right, clear the error message. """
        if self.prev_bottom != self.default_bottom:
            self._bottom_line(self.default_bottom)

    @staticmethod
    def get_next_key(key):
        """ Returns the next ASCII char after the given;
            e.g., 'b' = get_next_key('a')
        """
        # old_int = ord(key[0])
        # next_int = old_int + 1
        # next_chr = chr(next_int)
        # next_str = str(next_chr)
        # return next_str
        return str(chr(1+ord(key[0])))
    def get_command(self, obj):
        """ Return the 'command' from the returned Namespace which
            by convention is the 'extra', the 'paload' if a string,
            or the prompt ... in that priority order
        """
        if obj.extra:
            return obj.extra
        if isinstance(obj.payload, str):
            return obj.payload
        return obj.prompt

    def prompt(self):
        """ The simplified external entry point which prompts the user
        to select and choose a menu entry and returns the key  """
        return self.get_prompt_obj().key

    def get_prompt_obj(self):
        """ The external entry point which prompts the user
        to select and choose a menu entry an returns a namespace  """
        while True:
            try:
                self._prompt_mode(enable=True)
                ans = ''
                while True:
                    key = self._get_char()
                    ans += key
                    if key.isalnum() or key in ('\r', '\n', ' '):
                        break
                if ans in ('\033[A', '\033[C'):
                    self._restore_default_bottom()
                    self._move(-1)
                elif ans in ('\033[B', '\033[C'):
                    self._restore_default_bottom()
                    self._move(1)
                elif ans in ('\r', '\n', ' '):
                    if self.selected < len(self.lines):
                        self._restore_default_bottom()
                        picked = list(self.prompts.keys())[self.selected]
                        self._prompt_mode(enable=False)
                        # self.finish(f'\n\nrunning {repr(picked)}')
                        self._finish()
                        return self.objects[picked]
                    self._bottom_line('invalid(no selection); pick again')
                elif ans in self.prompts:
                    self._restore_default_bottom()
                    idx = list(self.prompts.keys()).index(ans)
                    self._move(idx - self.selected)
                else:
                    self._bottom_line(f'invalid({repr(ans)}); pick again')
            finally:
                self._prompt_mode(enable=False)

def runner(_): # def runner(argv):
    """ TBD """
    
    def simple_menu():
        default = 'A'
        while True:
            opts = {
                'q': 'Quit',
                'A': 'Set root password [suggested: "pw"]',
                'B': 'Set current user password [suggested: "pw"]',
                'u': 'Update Linux -- run after ChromeOS update (at least)',
                'r': 'Refresh Icons -- fix cases of icons becoming lost',
            }
            menu = Menu(opts, default=default, title='Simple Menu')
            key = menu.prompt()
            line = opts[key]
            print(f'\n===> To run: {line}\n')
            if 'Quit' in line:
                return 0
            default = menu.objects[key].next

    def multiline_menu():
        picked = 'A'
        while True:
            opts = {
                'A': 'Set root password [suggested: "pw"]\n'
                     'sudo passwd',
                'B': 'Set current user password [suggested: "pw"]\n'
                     'passwd',
                'u': 'Update Linux -- run after ChromeOS update (at least)\n'
                     'sudo pacman -Syu',
                'r': 'Refresh Icons -- fix cases of icons becoming lost\n'
                     'fix passwords somehow',
                'x': 'exit\n'
                     '# return from menu loop',
            }
            menu = Menu(opts, picked, title='MultiLine Menu')
            picked = menu.get_prompt_obj()
            print(f'\n===> To run: {picked.extra}\n')
            if 'exit' in picked.prompt:
                return 0

    def mixed_menu():
        picked = None
        while True:
            opts = {
                'A': 'Set root password [suggested: "pw"]\n'
                     'sudo passwd',
                'B': 'Set current user password [suggested: "pw"]\n'
                     'passwd',
                'u': ('Update Linux -- run after ChromeOS update (at least)\n'
                     'sudo pacman -Syu', {'key1': 'val1', 'key2': 'val2'}),
                'r': ('Refresh Icons -- fix cases of icons becoming lost',
                     {'Key1': 'Val1', 'Key2': 'Val2'}),
                'q': 'Quit\n',
            }
            menu = Menu(opts, picked, title='Mixed Menu')
            picked = menu.get_prompt_obj()
            print(f'\n===> returned: {vars(picked)}')
            if 'Quit' in picked.prompt:
                return 0
    
    simple_menu()
    multiline_menu()
    mixed_menu()

if __name__ == '__main__':
    runner(sys.argv)